/*  Copyright 1997-1998 Regents of the University of California.
Permission is granted to use this software for any noncommercial purpose.
For commercial licensing contact the UCSD Technology Transfer Office.

UC MAKES NO WARRANTY, EXPRESS OR IMPLIED, IN CONNECTION WITH THIS SOFTWARE!

Written by Miller Puckette (msp@ucsd.edu)
*/

#include "m_pd.h"
#include <math.h>
#include <stdio.h>

#ifdef NT
#pragma warning (disable: 4305 4244)
#endif

/* ------------------------ bonk~ ----------------------------- */

#define NPOINTS 256
#define DEFPERIOD 128
#define DEFHITHRESH 50
#define DEFLOTHRESH 6
#define DEFMASKTIME 4
#define DEFMASKDECAY 0.7
#define DEFDEBOUNCEDECAY 0
#define DEFMINVEL 7

static t_class *bonk_class;

typedef struct _filterspec
{
    int s_npoints;
    float s_freq;
} t_filterspec;

static t_filterspec bonk_filters[] =
    {{256, 2}, {256, 4}, {256, 6}, {180, 6}, {128, 6},
    {90, 6}, {64, 6}, {45, 6}, {32, 6}, {23, 6}, {16, 6}};

#define NFILTERS (sizeof(bonk_filters)/sizeof(bonk_filters[0]))

typedef struct _filterkernel
{
    int k_onset;
    int k_npoints;
    float *k_stuff;
} t_filterkernel;

static t_filterkernel bonk_filterkernels[NFILTERS];


typedef struct _hist
{
    float h_power;
    float h_mask;
    float h_before;
    int h_countup;
} t_hist;

typedef struct template
{
    float t_amp[NFILTERS];
} t_template;

typedef struct _bonk
{
    t_object x_obj;
    t_outlet *x_cookedout;
    t_clock *x_clock;
    t_hist x_hist[NFILTERS];
    t_template *x_template;
	int x_down2;
    int x_ntemplate;
    float *x_inbuf;
    int x_infill;
    int x_countdown;
    int x_period;
    int x_willattack;
    float x_hithresh;
    float x_lothresh;
    int x_masktime;
    float x_maskdecay;
    int x_learn;
    double x_learndebounce;	/* used only in "learn" mode */
    int x_learncount;	    	/* countup for "learn" mode */
    float x_debouncedecay;
    float x_minvel; 	    	/* minimum velocity we output */
    float x_debouncevel;
} t_bonk;

static void bonk_tick(t_bonk *x);

static void *bonk_new(t_floatarg fperiod, t_floatarg fdown2)
{
    int i;
    t_hist *h;
    t_bonk *x = (t_bonk *)pd_new(bonk_class);
    float *fp;
    int period = fperiod;
    int down2 = (fdown2 != 0);
    
    outlet_new(&x->x_obj, gensym("list"));
    x->x_cookedout = outlet_new(&x->x_obj, gensym("list"));
    x->x_clock = clock_new(x, bonk_tick);

    for (i = 0, h = x->x_hist; i--; h++)
    	h->h_power = h->h_mask = h->h_before = 0, h->h_countup = 0;
    x->x_template = (t_template *)getbytes(0);
    x->x_ntemplate = 0;
    x->x_inbuf = (float *)getbytes(NPOINTS * sizeof(float));
    for (i = NPOINTS, fp = x->x_inbuf; i--; fp++) *fp = 0;
    x->x_infill = 0;
    x->x_countdown = 0;
    if (!period) period = NPOINTS/2;
    x->x_period = period;
    x->x_willattack = 0;
    x->x_down2 = down2;
    x->x_hithresh = DEFHITHRESH;
    x->x_lothresh = DEFLOTHRESH;
    x->x_masktime = DEFMASKTIME;
    x->x_maskdecay = DEFMASKDECAY;
    x->x_learn = 0;
    x->x_learndebounce = clock_getsystime();
    x->x_learncount = 0;
    x->x_debouncedecay = DEFDEBOUNCEDECAY;
    x->x_minvel = DEFMINVEL;
    x->x_debouncevel = 0;

    return (x);
}

static void bonk_tick(t_bonk *x)
{
    t_atom at[NFILTERS], *ap, at2[2];
    int i, j;
    t_hist *h;
    float powerout[NFILTERS], *pp, vel = 0;
    float *fp;
    t_template *tp;
    int nfit;
    for (i = NFILTERS, pp = powerout, h = x->x_hist; i--; pp++, h++)
    {
    	float pow = h->h_mask - h->h_before;
    	if (pow > 0) vel += (*pp = 35.f * qrsqrt(qrsqrt(pow)));
		else *pp = 0;
    }
    if (vel < x->x_debouncevel || vel < x->x_minvel) return;
    x->x_debouncevel = vel;
    if (x->x_learn)
    {
    	double lasttime = x->x_learndebounce;
    	double msec = clock_gettimesince(lasttime);
    	if (!(x->x_ntemplate) || (msec > 200))
    	{
    	    int countup = x->x_learncount, ntemplate = x->x_ntemplate;
    	    	/* normalize to 100  */
    	    float norm;
    	    for (i = NFILTERS, norm = 0, pp = powerout;
    	    	i--; pp++) norm += *pp * *pp;
    	    if (norm < 1.0e-15) norm = 1.0e-15;
    	    norm = 100.f * qrsqrt(norm);
    	    	/* check if this is the first strike for a new template */
    	    if (!countup)
    	    {
    	    	int oldn = ntemplate;
    	    	x->x_ntemplate = ntemplate = oldn + 1;
    	    	x->x_template = (t_template *)t_resizebytes(x->x_template,
    	    	    oldn * sizeof(x->x_template[0]),
    	    	    ntemplate * sizeof(x->x_template[0]));
    	    	for (i = NFILTERS, fp = x->x_template[oldn].t_amp,
    	    	    pp = powerout;  i--; fp++, pp++)
    	    	    	*fp = *pp * norm;
    	    }
    	    else
    	    {
    	    	for (i = NFILTERS, fp = x->x_template[ntemplate-1].t_amp,
    	    	    pp = powerout;  i--; fp++, pp++)
    	    	    	*fp = (countup * *fp + *pp * norm) /(countup + 1.0f);
    	    }
    	    countup++;
    	    if (countup == x->x_learn) countup = 0;
    	    x->x_learncount = countup;
    	}
    	else return;
    }
    x->x_learndebounce = clock_getsystime();
    if (x->x_ntemplate)
    {
#if 0
    	    /*this is a more sophisticated test that seems to work no
    	    	better than the straight dot-product test below: */
    	float bestfit = 0;
    	nfit = -1;
    	for (i = 0, tp = x->x_template; i < x->x_ntemplate; tp++, i++)
    	{
    	    float dotprod = 0, wantsqnorm = 0, gotsqnorm = 0, score;
    	    for (j = NFILTERS, fp = tp->t_amp, pp = powerout, h = x->x_hist;
    	    	j--; fp++, pp++, h++)
	    {
		float want = *fp, got = *pp;
		float mask = h->h_mask, before = h->h_before;
		float weight;
		if (mask > before && before >= 0)
			weight = (mask - before)/mask;
		else weight = 0;
		dotprod += weight * want * got;
		wantsqnorm += weight * want * want;
		gotsqnorm += weight * got * got;
	    }
	    if (wantsqnorm > 0 && gotsqnorm > 0)
		    score = dotprod * dotprod / (wantsqnorm * gotsqnorm);
    	    else score = 0;
	    if (score > bestfit)
    	    {
    	    	bestfit = score;
    	    	nfit = i;
    	    }
    	}
    	if (nfit < 0)
	{
	    post("bonk: ouch");
	    nfit = 0;
	}
#endif
#if 1
    	float bestfit = 0;
    	nfit = -1;
    	for (i = 0, tp = x->x_template; i < x->x_ntemplate; tp++, i++)
    	{
    	    float dotprod = 0;
    	    for (j = NFILTERS, fp = tp->t_amp, pp = powerout;
    	    	j--; fp++, pp++) dotprod += *fp * *pp;
    	    if (dotprod > bestfit)
    	    {
    	    	bestfit = dotprod;
    	    	nfit = i;
    	    }
    	}
    	if (nfit < 0) bug("bonk_tick");
#endif
    }
    else nfit = 0;

    at2[0].a_type = A_FLOAT;
    at2[0].a_w.w_float = nfit;
    at2[1].a_type = A_FLOAT;
    at2[1].a_w.w_float = vel;
    outlet_list(x->x_cookedout, &s_list, 2, at2);
    for (i = 0, ap = at, pp = powerout, vel = 0; i < NFILTERS; i++, ap++, pp++)
    {
    	ap->a_type = A_FLOAT;
    	ap->a_w.w_float = *pp;
    }
    outlet_list(x->x_obj.ob_outlet, &s_list, NFILTERS, at);
}

static void bonk_doit(t_bonk *x)
{
    int i, j;
    t_filterkernel *k;
    t_hist *h;
    float growth = 0;
    
    for (i = NFILTERS, k = bonk_filterkernels, h = x->x_hist;
		i--; k++, h++)
    {
    	float rsum = 0, isum = 0, *fp1, *fp2, power, maskpow = h->h_mask;
    	int countup = h->h_countup;
    	for (fp1 = x->x_inbuf + k->k_onset,
			fp2 = k->k_stuff, j = k->k_npoints; j--;)
    	{
    	    float g = *fp1++;
    	    rsum += g * *fp2++;
    	    isum += g * *fp2++;
    	}
    	power = rsum * rsum + isum * isum;

	if (!x->x_willattack) h->h_before = maskpow;

    	if (power > maskpow)
    	    growth += power/(maskpow + 1.0e-15) - 1.f;

    	if (!x->x_willattack && countup >= x->x_masktime)
	    maskpow *= x->x_maskdecay;

    	if (power > maskpow)
    	{
    	    maskpow = power;
    	    countup = 0;
    	}
    	countup++;
    	h->h_countup = countup;
    	h->h_mask = maskpow;
    	h->h_power = power;
    }

    if (x->x_willattack > 4)
    {
    	    /* if it takes more than 4 analyses for the energy to stop growing,
    	    forget it; we would rather miss the note than report it late. */
    	/* post("soft attack"); */
    	x->x_willattack = 0;
    }
    else if (x->x_willattack)
    {
    	if (growth < x->x_lothresh)
    	{
    	    clock_delay(x->x_clock, 0);
    	    x->x_willattack = 0;
    	}
    	else x->x_willattack++;
    }
    else if (growth > x->x_hithresh)
    {
	x->x_willattack = 1;
	for (i = NFILTERS, h = x->x_hist; i--; h++)
	    h->h_mask = h->h_power, h->h_countup = 0;
    }
    
    x->x_debouncevel *= x->x_debouncedecay;
    
    	/* shift the input buffer and update counters */
    if (x->x_period > NPOINTS) x->x_countdown = x->x_period - NPOINTS;
    else x->x_countdown = 0;
    if (x->x_period < NPOINTS)
    {
    	int overlap = NPOINTS - x->x_period;
    	float *fp1, *fp2;
    	for (i = overlap, fp1 = x->x_inbuf, fp2 = fp1 + x->x_period;
    	    i--;) *fp1++ = *fp2++;
    	x->x_infill = overlap;
    }
    else x->x_infill = 0;	
}

static t_int *bonk_perform(t_int *w)
{
    t_float *in1 = (t_float *)(w[1]);
    t_bonk *x = (t_bonk *)(w[2]);
    int n = (int)(w[3]);
    int down2 = x->x_down2;

    if (x->x_countdown > 0) x->x_countdown -= (down2? (n>>1) : n);
    else
    {
    	int infill = x->x_infill;
    	float *fp = x->x_inbuf + infill;
	if (down2)
	{
	    int nreal = n>>1;
    	    infill += nreal;
    	    while (nreal--)
	    {
		*fp++ = *in1;
		in1 += 2;
	    }
	}
	else
	{
    	    infill += n;
    	    while (n--) *fp++ = *in1++;
	}
    	x->x_infill = infill;   
    	if (infill == NPOINTS) bonk_doit(x);
    }
    return (w+4);
}

static void bonk_dsp(t_bonk *x, t_signal **sp)
{
    dsp_add(bonk_perform, 3, sp[0]->s_vec, x, sp[0]->s_n);
}

static void bonk_thresh(t_bonk *x, t_floatarg f1, t_floatarg f2)
{
    if (f1 > f2)
    	post("bonk: warning: low threshold greater than hi threshold");
    x->x_lothresh = f1;
    x->x_hithresh = f2;
}

static void bonk_mask(t_bonk *x, t_floatarg f1, t_floatarg f2)
{
    int ticks = f1;
    if (ticks < 0) ticks = 0;
    if (f2 < 0) f2 = 0;
    else if (f2 > 1) f2 = 1;
    x->x_masktime = ticks;
    x->x_maskdecay = f2;
}

static void bonk_debounce(t_bonk *x, t_floatarg f1)
{
    if (f1 < 0) f1 = 0;
    else if (f1 > 1) f1 = 1;
    x->x_debouncedecay = f1;
}

static void bonk_minvel(t_bonk *x, t_floatarg f)
{
    if (f < 0) f = 0; 
    x->x_minvel = f;
}

static void bonk_print(t_bonk *x, t_floatarg f)
{
    int i;
    post("thresh %f %f", x->x_lothresh, x->x_hithresh);
    post("mask %d %f", x->x_masktime, x->x_maskdecay);
    post("debounce %f", x->x_debouncedecay);
    post("minvel %f", x->x_minvel);
    if (x->x_ntemplate)
    {
    	post("templates:");
    	for (i = 0; i < x->x_ntemplate; i++)
    	    post("%2d \
%5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f", i,
    	    x->x_template[i].t_amp[0],
    	    x->x_template[i].t_amp[1],
    	    x->x_template[i].t_amp[2],
    	    x->x_template[i].t_amp[3],
    	    x->x_template[i].t_amp[4],
    	    x->x_template[i].t_amp[5],
    	    x->x_template[i].t_amp[6],
    	    x->x_template[i].t_amp[7],
    	    x->x_template[i].t_amp[8],
    	    x->x_template[i].t_amp[9],
    	    x->x_template[i].t_amp[10]);
    }
    else post("no templates");
    if (x->x_learn) post("learn mode");
    if (f != 0)
    {
    	int i;
    	t_hist *h;
    	for (i = NFILTERS, h = x->x_hist; i--; h++)
    	    post("pow %f mask %f before %f count %d",
    	    	h->h_power, h->h_mask, h->h_before, h->h_countup);
    }
}

static void bonk_learn(t_bonk *x, t_floatarg f)
{
    int n = f;
    if (n < 0) n = 0;
    if (n)
    {
    	x->x_template = (t_template *)t_resizebytes(x->x_template,
    	    x->x_ntemplate * sizeof(x->x_template[0]), 0);
    	x->x_ntemplate = 0;
    }
    x->x_learn = n;
    x->x_learncount = 0;
}

static void bonk_forget(t_bonk *x)
{
    int ntemplate = x->x_ntemplate, newn = ntemplate-1;
    if (ntemplate)
    {
    	x->x_template = (t_template *)t_resizebytes(x->x_template,
    	    x->x_ntemplate * sizeof(x->x_template[0]),
    	    newn * sizeof(x->x_template[0]));
    	x->x_ntemplate = newn;
    }
    x->x_learncount = 0;
}

static void bonk_bang(t_bonk *x)
{
    t_atom at[NFILTERS];
    int i;
    for (i = 0; i < NFILTERS; i++)
    {
    	at[i].a_type = A_FLOAT;
    	at[i].a_w.w_float = 100 * qrsqrt(qrsqrt(x->x_hist[i].h_power));
    }
    outlet_list(x->x_obj.ob_outlet, &s_list, NFILTERS, at);
}

static void bonk_setupkernels(void)
{
    int i, j;
    float *fp;
    for (i = 0; i < NFILTERS; i++)
    {
    	int npoints = bonk_filters[i].s_npoints;
    	float freq = bonk_filters[i].s_freq;
    	float normalize = 2.8284f / npoints;
    	float phaseinc = (2.f * 3.14159f) / npoints;
    	bonk_filterkernels[i].k_npoints = npoints;
    	bonk_filterkernels[i].k_onset = (NPOINTS/2) - (npoints/2);
    	bonk_filterkernels[i].k_stuff =
    	    (float *)getbytes(2 * sizeof(float) * npoints);
    	for (fp = bonk_filterkernels[i].k_stuff, j = npoints; j--;)
    	{
    	    float phase = j * phaseinc;
    	    float window = normalize * (0.5f - 0.5f * cos(phase));
    	    *fp++ = window * cos(freq * phase);
    	    *fp++ = window * sin(freq * phase);
    	}
    }
}

static void bonk_read(t_bonk *x, t_symbol *s)
{
	FILE *fd = fopen(s->s_name, "r");
    float vec[NFILTERS];
    int i, ntemplate = 0;
    float *fp, *fp2;
    if (!fd)
    {
	post("%s: open failed", s->s_name);
	return;
    }
    x->x_template = (t_template *)t_resizebytes(x->x_template, 
	x->x_ntemplate * sizeof(t_template), 0);
    while (1)
    {
	for (i = NFILTERS, fp = vec; i--; fp++)
	    if (fscanf(fd, "%f", fp) < 1) goto nomore;
	x->x_template = (t_template *)t_resizebytes(x->x_template,
	    ntemplate * sizeof(t_template),
	    (ntemplate + 1) * sizeof(t_template));
	for (i = NFILTERS, fp = vec,
	    fp2 = x->x_template[ntemplate].t_amp; i--;)
		*fp2++ = *fp++;
	ntemplate++;
    }
nomore:
    post("bonk: read %d templates\n", ntemplate);
    x->x_ntemplate = ntemplate;
    fclose(fd);
}

static void bonk_write(t_bonk *x, t_symbol *s)
{
    FILE *fd = fopen(s->s_name, "w");
    int i, ntemplate = x->x_ntemplate;
    t_template *tp = x->x_template;
    float *fp;
    if (!fd)
    {
	post("%s: couldn't create", s->s_name);
	return;
    }
    for (; ntemplate--; tp++)
    {
	for (i = NFILTERS, fp = tp->t_amp; i--; fp++)
		fprintf(fd, "%6.2f ", *fp);
	fprintf(fd, "\n");
    }
    post("bonk: wrote %d templates\n", x->x_ntemplate);
    fclose(fd);
}

static void bonk_free(t_bonk *x)
{
    clock_free(x->x_clock);
    freebytes(x->x_inbuf, NPOINTS * sizeof(float));
}

void bonk_setup(void)
{
    bonk_class = class_new(gensym("bonk"), (t_newmethod)bonk_new, 0,
    	sizeof(t_bonk), 0, A_DEFFLOAT, A_DEFFLOAT, 0);
    class_addmethod(bonk_class, nullfn, gensym("signal"), 0);
    class_addmethod(bonk_class, bonk_dsp, gensym("dsp"), 0);
    class_addbang(bonk_class, bonk_bang);
    class_addmethod(bonk_class, (t_method)bonk_learn, gensym("learn"),
    	A_FLOAT, 0);
    class_addmethod(bonk_class, (t_method)bonk_forget, gensym("forget"), 0);
    class_addmethod(bonk_class, (t_method)bonk_thresh, gensym("thresh"),
    	A_FLOAT, A_FLOAT, 0);
    class_addmethod(bonk_class, (t_method)bonk_mask, gensym("mask"),
    	A_FLOAT, A_FLOAT, 0);
    class_addmethod(bonk_class, (t_method)bonk_debounce, gensym("debounce"),
    	A_FLOAT, 0);
    class_addmethod(bonk_class, (t_method)bonk_minvel, gensym("minvel"),
    	A_FLOAT, 0);
    class_addmethod(bonk_class, (t_method)bonk_print, gensym("print"),
    	A_DEFFLOAT, 0);
    class_addmethod(bonk_class, bonk_read, gensym("read"),
    	A_SYMBOL, 0);
    class_addmethod(bonk_class, bonk_write, gensym("write"),
    	A_SYMBOL, 0);
    bonk_setupkernels();
    post("bonk version 1.0");
}
